/* glmrunner_v023.js
GLM computations 
For BrainVoyager 21 and higher

- v0.2.3: changed 'BrainVoyagerQX' object to 'BrainVoyager' for BrainVoyager 21 (08-04-19)
- v0.2.3: fix in generating MDM (28-08-15)
- v0.2.2: added loading of stimulation protocols
- v0.2.1: updated headers to BrainVoyager QX 2.8.0; created VMPs are now NR-VMP instead of AR-VMP
- v0.2: also create optionally one VMP per single-study GLM, optionally with contrast specification
- v0.1: create multiple single-study GLMs and optionally an *.mdm, multi-study GLM and RFX GLM
HB BI 17-07-12, updates c++ code Mrch 15
*/

// widget names
var btnLoadFilenames, lblVMRname, btnLoadVMR, chbCorrectSerCor, chbVTC, chbFMR, spbNrGLMs;
var chbMultiSubjGLM, chbRFXGLM, chbSeparateStudies, chbSeparateSubjects, rbtnZtrfBaseline, rbtnZtrf, rbtnPSC;
var rbtnReadTextfile, rbtnSelectViaDlg, btnStart;
var chbSaveVMP, chbSpecContrast, leContrastName, leContrastStr;

// constants
const TEXTFILE_FOR_SSGLM        = 1;
const TEXTFILE_FOR_POST_SSGLM   = 2;
const FILETYPE_UNKNOWN          = 0;
const FILETYPE_FMR              = 1;
const FILETYPE_VTC              = 2;

// communication between user interface and script
var scriptObj = new Object; 
scriptObj.scriptNameForUser     = "glmrunner"; 
scriptObj.dialogFileName        = "glmrunner_v023.ui";
scriptObj.dialogAccessName      = "glmDlg"; 

// script global variable
var projfilenames; 

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

scriptObj.startProcessing = function() {

 //   var success = this.getDesMatAndFuncFiles();
    var success = (projfilenames.length > 0);
    if (success) {
        Compute_GLMs(); 
    }
    this.glmDlg.close();
}

function Compute_GLMs() {
    
	var counter, doc, funcfile, desmatfile, prtfile;
    
    // always run single-subject
    var nroffiles = (projfilenames.length);
    var filetype = (chbVTC.checked ? FILETYPE_VTC : FILETYPE_FMR);
    var anatname = lblVMRname.text;
    if ((anatname.length < 1) && FILETYPE_VTC) BrainVoyager.PrintToLog("Please note: this option requires an anatomical file.");
    
    BrainVoyager.PrintToLog("Start calculating " + nroffiles + " single-study GLMs..."); 
    for (counter = 0; counter < nroffiles; counter++) {	 
    
        prtfile = projfilenames[counter][0];
        funcfile = projfilenames[counter][1];
        BrainVoyager.PrintToLog("Functional file: " + funcfile);                   
        desmatfile = projfilenames[counter][2];
            
        if (filetype == FILETYPE_FMR) {		
                BrainVoyager.PrintToLog("Opening " + funcfile + "..."); // functional data (*.fmr + *.stc)
                doc = BrainVoyager.OpenDocument(funcfile); 
        } else if (filetype == FILETYPE_VTC){
                BrainVoyager.PrintToLog("Opening " + anatname + "..."); // anatomical data (*.vmr)
                doc = BrainVoyager.OpenDocument(anatname); 
                BrainVoyager.PrintToLog("Linking " + funcfile + "..."); // functional data (*.vtc)
                doc.LinkVTC(funcfile); 
        }
        
        //>> clear stim prt + desmat, just to be sure?
        
        BrainVoyager.PrintToLog("Linking " + prtfile + "...");
        doc.LinkStimulationProtocol(prtfile);
        BrainVoyager.PrintToLog("Loading " + desmatfile + "...");
        doc.LoadSingleStudyGLMDesignMatrix(desmatfile);	// design matrix (*.sdm)
        doc.CorrectForSerialCorrelations = chbCorrectSerCor.checked;
        BrainVoyager.PrintToLog("Correct serial correlations: " + (doc.CorrectForSerialCorrelations ? "yes" : "no"));
        doc.ComputeSingleStudyGLM();
        
        var glmname = removeextension(funcfile) + ".glm";		
        BrainVoyager.PrintToLog("Saving " + glmname + "...");	
        doc.SaveGLM(glmname); 

        if (chbSaveVMP.checked) {
        
            var contrastname, contrast;
            doc.LoadGLM(glmname); 
            if (chbSpecContrast.checked) {
                doc.ClearContrasts(); 
                contrastname = leContrastName.text;
                BrainVoyager.PrintToLog("Contrast name: " + contrastname);
                doc.AddContrast(contrastname);  
                contrast = leContrastStr.text;
                BrainVoyager.PrintToLog("Set contrast string: " + contrast);
                doc.SetContrastString(contrast);
            }
            doc.ShowGLM(); 
            BrainVoyager.SetPluginStringParameter("Command", "CreateVMP"); 
            if (chbSpecContrast.checked) {
                BrainVoyager.SetPluginStringParameter("VMPMapName", removeextension(glmname) + "_" + contrastname + ".vmp");             
            } else {
                BrainVoyager.SetPluginStringParameter("VMPMapName", removeextension(glmname) + ".vmp"); 
            }
            BrainVoyager.ExecutePlugin();  
        }
        
/*        var filenamesrf = BrainVoyager.GetPluginStringParameter("FilenameSrf");
        doc.LoadMesh(filenamesrf);
        BrainVoyager.SetPluginStringParameter("Command", "CreateSMP"); 
        BrainVoyager.PrintToLog("Try to create SMP...");
        BrainVoyager.ExecutePlugin();         
  */      
        
        doc.Close();
    } // end counter		  
	
	if  ((chbMultiSubjGLM.checked) || (chbRFXGLM.checked)) {
		
         var anatname = lblVMRname.text;
         var mdmname = Generate_MDM_file(projfilenames, anatname); // In BrainVoyager QX 2.1 only for VTCs 
        
		 BrainVoyager.PrintToLog("Opening " + anatname + "..."); // anatomical data (*.vmr)
		 doc = BrainVoyager.OpenDocument(anatname); 
		 doc.LoadMultiStudyGLMDefinitionFile(mdmname);
		 doc.PSCTransformStudies 			= rbtnPSC.checked;
		 doc.ZTransformStudies 				= rbtnZtrf.checked;
		 doc.ZTransformStudiesBaselineOnly 	= rbtnZtrfBaseline.checked;
		 doc.SeparationOfSubjectPredictors 	= chbSeparateSubjects.checked;
		 doc.SeparationOfStudyPredictors 	= chbSeparateStudies.checked;
		 BrainVoyager.PrintToLog("Predictor separation: " + (doc.SeparationOfSubjectPredictors ? "per subject" : (doc.SeparationOfStudyPredictors ? "per run" : "no")));
		 BrainVoyager.PrintToLog("Data normalization: " + (doc.ZTransformStudies ? "z-transform" : (doc.ZTransformStudiesBaselineOnly ? "z-transform (baseline)" : 
										(doc.PSCTransformStudies ? "percent signal change " : "not set"))));
		 
         if (chbMultiSubjGLM.checked) {
            doc.ComputeMultiStudyGLM();
            var glmname = removeextension(mdmname) + ".glm";		
            BrainVoyager.PrintToLog("Saving " + glmname + "...");	
            doc.SaveGLM(glmname);
            doc.ShowGLM();
         }
         if (chbRFXGLM.checked) {        
            doc.ComputeRFXGLM();
            var rfxglmname = removeextension(mdmname) + "_rfx.glm";		
            BrainVoyager.PrintToLog("Saving " + rfxglmname + "...");	
            doc.SaveGLM(rfxglmname);
       //     doc.ShowGLM(); // bug? (crashes)
         }
	}
}

function Generate_MDM_file(projfilenames, anatname) {
    
    var mdmname;
    BrainVoyager.ShowLogTab();
	BrainVoyager.PrintToLog("Start creating MDM...\nOpening anatomical file: " + anatname);
    var nroffiles = projfilenames.length;
    var anat = BrainVoyager.OpenDocument(anatname);
	if (anat == undefined) {
		BrainVoyager.PrintToLog("Invalid anatomical filename.");
	} else {
        anat.ClearMultiStudyGLMDefinition();
        for (counter = 0; counter < nroffiles; counter++) {
            BrainVoyager.PrintToLog("Adding:\n" + projfilenames[counter][1] + "\n" + projfilenames[counter][2]);
            anat.AddStudyAndDesignMatrix(projfilenames[counter][1], projfilenames[counter][2]); // fix 28-08-15 from PRT to SDM
        }
		mdmname = removeextension(projfilenames[0][1]) + "_" + nroffiles + "studies.mdm";// modify *.sdm name
		BrainVoyager.PrintToLog("Name for MDM file: " + mdmname);
		anat.SaveMultiStudyGLMDefinitionFile(mdmname);
    }
    return mdmname;
}

function removeextension(filename) {
	var extlength = getextensionlength(filename);
	var name = "";
	for (i=0; i<(filename.length-extlength); i++) {
		letter = filename[i];
		name += letter;
	}	
	return name;
}

function getextensionlength(filename) {
	var extensionlength, letter, dotpos;
	for (i=filename.length; i>0; i--) {
		letter = filename[i];
		if (letter == ".") {
			dotpos = i;					
			break;
		}	
	}
	extensionlength = filename.length - dotpos;
	return extensionlength;
}

scriptObj.getFile = function(filetype) {

    var success = true;
    BrainVoyager.SetPluginStringParameter("FileType", filetype);  
    BrainVoyager.SetPluginStringParameter("Command", "GetFileName"); 
    BrainVoyager.ExecutePlugin(); 
    var filename = BrainVoyager.GetPluginStringParameter("Filename");
    return filename;
}

scriptObj.getAnatFile = function() {

	BrainVoyager.PrintToLog("Get anatomical file (*.vmr)..."); 
    var filetype = "Anatomical file (*.vmr)";
    var filename = this.getFile(filetype);
    if (filename.length > 0) {
        lblVMRname.text = filename;
        BrainVoyager.PrintToLog("Name: " + filename);
    }
}

scriptObj.getDesMatAndFuncFiles = function() {

    var success = true;
	BrainVoyager.PrintToLog("Get files..."); 
    var filetype = "*.*";
    if (chbFMR.checked) {
        filetype = "*.fmr";
    } else {
        filetype = "*.vtc";
    } 
    
    // select files manually
    projfilenames = new Array();
    if (rbtnSelectViaDlg.checked == 1) {
    
        var nroffiles = spbNrGLMs.value;
        if (nroffiles < 1) BrainVoyager.PrintToLog("Please note: number of selected files is 0.");
        BrainVoyager.PrintToLog("Start selecting: " + nroffiles + " files..."); 
        var filecounter;
        // loop over indicated number of files
        for (filecounter = 0; filecounter < nroffiles; filecounter++) {
            
            var groupnames = new Array(3);
            var prtname = this.getFile("*.prt");
            if (prtname.length > 0) groupnames[0] = prtname;
            var funcfilename = this.getFile(filetype);
            if (funcfilename.length > 0) groupnames[1] = funcfilename;
            var desmatname = this.getFile("*.sdm");
            if (desmatname.length > 0) groupnames[2] = desmatname;
            projfilenames.push(groupnames);
            
        }
        this.writeToTextFile(projfilenames, TEXTFILE_FOR_SSGLM, "glm");// stage, targetfilenames
        BrainVoyager.PrintToLog("Files selected. If all parameters have been set, the Start button can be clicked.");
        
    // read filenames from textfile    
    } else if (rbtnReadTextfile.checked == 1) {
    
        BrainVoyager.SetPluginStringParameter("Command", "GetTextFileName");
        BrainVoyager.PrintToLog("Select text file with source filenames..."); 
        BrainVoyager.ExecutePlugin();
        if (this.noErrors()) {
            var textfilename = BrainVoyager.GetPluginStringParameter("Filename");
            BrainVoyager.PrintToLog("Selected text filename: " + textfilename);
            this.readTextFileIntoProjectNamesArray(textfilename, TEXTFILE_FOR_SSGLM); 
        } else {
            success = false;
        }
    }    
    return success;
}

scriptObj.writeToTextFile = function(projfilenames, kindoffiles, stage) {//, stage, targetfilenames) 
    
    var filecounter;
    var success;
    
    // write the textfile to the path of the first file
    var filetype = projfilenames[0][0].substring((projfilenames[0][0].length-3), projfilenames[0][0].length);
    var name = String(this.getPath(projfilenames[0][0]) + "filenames_" + stage + "_" + filetype.toUpperCase() + ".txt");
    var filenamesfile = new QFile(name);     
    filenamesfile.open(new QIODevice.OpenMode(QIODevice.WriteOnly)); 
    var textstr = new QTextStream(filenamesfile); 
    textstr.writeDouble(projfilenames.length);
    textstr.writeString("\n");
    if (kindoffiles == TEXTFILE_FOR_SSGLM) {
        BrainVoyager.PrintToLog("Write design matrix and functional names...");
        for (filecounter = 0; filecounter < projfilenames.length; filecounter++) {
            textstr.writeString(projfilenames[filecounter][0] + "\n"); // stimulation protocol (PRT)
            textstr.writeString(projfilenames[filecounter][1] + "\n"); // source file (VTC or FMR)
            textstr.writeString(projfilenames[filecounter][2] + "\n"); // design matrix file (SDM)
        }
    } else {
        BrainVoyager.PrintToLog("Kind of files to write to textfile unknown.");
        success = false;
    }
    filenamesfile.close();
    BrainVoyager.PrintToLog("Wrote " + name + ".");
    return success;
}

scriptObj.readTextFileIntoProjectNamesArray = function(textfilename, kindoffiles) {

    var filecounter;    
    var success = true;
    BrainVoyager.PrintToLog("Start reading: " + textfilename);
    var name = String(textfilename); // circumvention of problem with implicit strings
    var filenamesfile = new QFile(name);     
    filenamesfile.open(new QIODevice.OpenMode(QIODevice.ReadOnly));
    var textstr = new QTextStream(filenamesfile); 
    var nroffiles = parseInt(textstr.readLine());
    BrainVoyager.PrintToLog("Number of projects: " + nroffiles);
    BrainVoyager.PrintToLog("Kind of projects: " + ((kindoffiles == TEXTFILE_FOR_SSGLM) ? "single-subject GLM" : "unknown")); 

    if (kindoffiles == TEXTFILE_FOR_SSGLM) {

        projfilenames = new Array(nroffiles);
        if (success) {
            for (filecounter = 0; filecounter < nroffiles; filecounter++) {
                
                BrainVoyager.PrintToLog("Files for GLM " + (filecounter+1) + ":");
                var groupnames = new Array(3);
                // *.prt
                var filename = textstr.readLine();
                BrainVoyager.PrintToLog(filename);
                groupnames[0] = filename;
                // *.vtc or *.fmr
                var filename = textstr.readLine();
                BrainVoyager.PrintToLog(filename); 
                groupnames[1] = filename;
                // *.sdm
                var filename = textstr.readLine();
                BrainVoyager.PrintToLog(filename); 
                groupnames[2] = filename;
                // save names in global array
                projfilenames[filecounter] = groupnames;                
            } 
        } 
        
        // set rest of parameters
        spbNrGLMs.value = nroffiles;
        if (nroffiles > 0) {
 //         var filetype = this.getFileType(projfilenames[0][0]); // 08-04-19
            var filetype = this.getFileType(projfilenames[0][1]);
            if (filetype == FILETYPE_VTC) {
                chbVTC.checked = true;
                chbMultiSubjGLM.enabled = true;
                chbRFXGLM.enabled = true;                
            } else if (filetype == FILETYPE_FMR) {
                chbFMR.checked = true;
                chbMultiSubjGLM.enabled = false;
                chbRFXGLM.enabled = false;
            } else {
                BrainVoyager.PrintToLog("Filetype of functional file unknown.");
            }
        }
        
    } else {
        BrainVoyager.PrintToLog("Kind of files unknown.");
    }
    filenamesfile.close();
    return success;
}

scriptObj.getPath = function(filename) { 

	var start = 0; // start searching from begin 
	var last = filename.lastIndexOf("/");
	var path = filename.substring(start, (last+1)); // now return path without filename
	return path;
}

scriptObj.noErrors = function() {

    var noerror;
    var errornr = BrainVoyager.GetPluginIntParameter("ErrorNr");
    if (errornr != -1) {
        noerror = true;
    } else {
        var errormsg = BrainVoyager.GetPluginStringParameter("ErrorMsg");
        BrainVoyager.PrintToLog(errormsg);
        noerror = false;        
    }
    return noerror;
}  

/*scriptObj.getFileTypeFromArray = function(projfilenames, row, column) {

    var filetype = projfilenames[row][column].substring((projfilenames[row][column].length-3), projfilenames[row][column].length);
    return filetype;
}
*/
scriptObj.getFileType = function(filename) {


var filetype;
    var filetypestr = filename.substring((filename.length-3), filename.length);
    if (filetypestr == "vtc") {
        filetype = FILETYPE_VTC;
    } else if (filetypestr == "fmr") {
        filetype = FILETYPE_FMR;
    } else {
        filetype = FILETYPE_UNKNOWN;
    }
    return filetype;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

scriptObj.initDlg = function() {

    // find widgets
    spbNrGLMs               = BrainVoyager.FindChild("spbNrGLMs");
    btnLoadFilenames        = BrainVoyager.FindChild("btnLoadFilenames");
    lblVMRname              = BrainVoyager.FindChild("lblVMRname");
    btnLoadVMR              = BrainVoyager.FindChild("btnLoadVMR");
    rbtnReadTextfile        = BrainVoyager.FindChild("rbtnReadTextfile");
    rbtnSelectViaDlg        = BrainVoyager.FindChild("rbtnSelectViaDlg");
    chbCorrectSerCor        = BrainVoyager.FindChild("chbCorrectSerCor");
    chbVTC                  = BrainVoyager.FindChild("chbVTC");
    chbFMR                  = BrainVoyager.FindChild("chbFMR");
    chbMultiSubjGLM         = BrainVoyager.FindChild("chbMultiSubjGLM");
    chbRFXGLM               = BrainVoyager.FindChild("chbRFXGLM");
    chbSeparateStudies      = BrainVoyager.FindChild("chbSeparateStudies");
    chbSeparateSubjects     = BrainVoyager.FindChild("chbSeparateSubjects");
    rbtnZtrfBaseline        = BrainVoyager.FindChild("rbtnZtrfBaseline");
    rbtnZtrf                = BrainVoyager.FindChild("rbtnZtrf");
    rbtnPSC                 = BrainVoyager.FindChild("rbtnPSC");
    btnStart                = BrainVoyager.FindChild("btnStart");
    chbSaveVMP              = BrainVoyager.FindChild("chbSaveVMP");
    chbSpecContrast         = BrainVoyager.FindChild("chbSpecContrast");
    leContrastName          = BrainVoyager.FindChild("leContrastName");
    leContrastStr           = BrainVoyager.FindChild("leContrastStr");

    // connect widgets to script functions   
    btnLoadVMR.clicked.connect(this, this.getAnatFile);
    btnLoadFilenames.clicked.connect(this, this.getDesMatAndFuncFiles);
    btnStart.clicked.connect(this, this.startProcessing);
 
}

returnScriptObj = function()
{
   return scriptObj; // it is important to return script object containing properties and functions to be called from BrainVoyager QX
}
returnScriptObj();
